import logging
import os
import platform
from datetime import datetime
from zipfile import ZipFile

import click
import numpy as np
import pandas as pd
import wget
from midas.util.runtime_config import RuntimeConfig

from . import meta

LOG = logging.getLogger(__name__)

if platform.system() == "Windows" or platform.system() == "Darwin":
    import ssl

    ssl._create_default_https_context = ssl._create_unverified_context


def download_weather(data_path, tmp_path, force):
    """Download and convert the weather datasets.

    The weather data is downloaded from https://opendata.dwd.de,
    the selected weather station is located in Bremen.

    At the beginning of every new year, the data from the previous
    year is added to the dataset. Unfortunately, the download link
    changes at the same time, now including the latest year of data.

    To prevent a failure every year, the year value in the download
    link is increased, but that may break at anytime if DWD decides to
    change the download links in any other way.

    The year is included in the 'post_fix' key of the runtime config.

    """
    LOG.info("Preparing weather datasets...")

    # We allow multiple datasets here
    for config in RuntimeConfig().data["weather"]:
        output_path = os.path.abspath(os.path.join(data_path, config["name"]))

        if os.path.exists(output_path):
            LOG.debug("Found existing dataset at %s.", output_path)
            if not force:
                continue
            else:
                LOG.debug("Downloading weather data anyways...")
        else:
            LOG.debug("No dataset found. Start downloading weather data ...")

        base_url = config["base_url"]
        post_fix = config["post_fix"]
        year = int(post_fix[:4])
        for url in [
            "pressure_url",
            "solar_url",
            "air_url",
            "cloud_url",
            "sun_url",
            "wind_url",
        ]:
            for idx in range(10):
                if url == "solar_url":
                    full_url = base_url + config[url]
                else:
                    # The other urls follow the same schema, including the
                    # latest year of data in the download link.
                    full_url = base_url + config[url] + f"{year}{post_fix[4:]}"

                fname = full_url.rsplit("/", 1)[-1]
                fpath = os.path.join(tmp_path, fname)
                if not os.path.exists(fpath):
                    try:
                        LOG.info("Start downloading '%s'...", full_url)
                        fpath = wget.download(full_url, out=tmp_path)
                        click.echo()
                        LOG.debug("Download complete.")
                        break
                    except Exception as err:
                        LOG.warning(
                            "Error during the download of file '%s': '%s'.\n"
                            "Probably the year has changed. Will try to fix "
                            "this automatically.",
                            full_url,
                            err,
                        )
                        fpath = None
                        if url != "solar_url":
                            year += 1
                        continue
                break

            if fpath is None:
                raise ValueError(
                    "Could not download weather data. Sorry for that. "
                    "This needs to be fixed manually :("
                )
            else:
                with ZipFile(os.path.join(tmp_path, fpath), "r") as zip_ref:
                    zip_ref.extractall(
                        os.path.join(tmp_path, url.split("_")[0])
                    )

        LOG.debug("Creating database...")
        # We start at 2009 because the solar dataset has no data before
        # that year.
        start_date = str(config.get("start_date", "2009-01-01 00:00:00"))
        data = pd.DataFrame(
            index=pd.date_range(
                start=start_date,
                end=f"{year}-12-31 23:00:00",
                tz="Europe/Berlin",
                freq="h",
            )
        )

        data = load_data(tmp_path, "pressure", data, year, start_date)
        data = load_data(tmp_path, "air", data, year, start_date)
        data = load_data(tmp_path, "solar", data, year, start_date)
        data = load_data(tmp_path, "wind", data, year, start_date)
        data = load_data(tmp_path, "cloud", data, year, start_date)
        data = load_data(tmp_path, "sun", data, year, start_date)

        data.index = data.index.tz_convert("UTC")
        # data.to_hdf(output_path, "weather", "w")
        data.to_csv(output_path, index_label="timestamp")
        LOG.info("Successfully created database for weather data.")


# Horizontal solar radiation is provided as hourly sum in Joule/cm^2
# (i.e., correct would be Joule/s/cm^2 * 3600s), but we want Watt/m^2
# for our PV models. Since 1*W = 1*J/s we first need to get back to
# J/s by dividing by 3600. Next, we want to convert from cm^2 to m^2,
# which is by multiplying with 0.0001, however, since cm^2 is in the
# divisor, we need to divide by that value (or multiply with the
# reciprocal). So the calculation we need to apply is
# 1 / (3.6*1e^3) * 1 / 1e^-4 = 1e^4 / (3.6*1e^3) = 1e^1 / 3.6
# which is equal to:
JOULE_TO_WATT = 10 / 3.6
DATA_COLS = {
    "air": [("TT_TU", meta.T_AIR, 1.0)],
    "solar": [
        ("FD_LBERG", meta.DI, JOULE_TO_WATT),
        ("FG_LBERG", meta.GHI, JOULE_TO_WATT),
    ],
    "wind": [("   F", meta.WIND, 1.0), ("   D", meta.WINDDIR, 1.0)],
    "cloud": [(" V_N", meta.CLOUDINESS, 12.5)],
    "sun": [("SD_SO", meta.SUN_HOURS, 1.0)],
    "pressure": [("   P", meta.PRESSURE, 1.0)],
}


def load_data(path, target, data, year, start_date):
    """Load data from a csv file and add them to a dataframe.

    Parameters
    ----------
    path: str
        The path of the folder containing a folder with csv files.
    target: str
        The name of the folder which contains csv files.
    data: pd.DataFrame
        The dataframe to which the content of the csv file will be
        added.
    year: int
        Since the year may change over the years, we pass it here.

    """
    fname = os.path.join(path, target)
    files = os.listdir(fname)
    data_file = [f for f in files if f.startswith("produkt")][0]
    fname = os.path.join(fname, data_file)

    if target == "solar":
        # We need a different parser for the solar dataset
        def parser(date):
            return datetime.strptime(date.split(":")[0], "%Y%m%d%H")

    else:

        def parser(date):
            return datetime.strptime(date, "%Y%m%d%H")

    # Read and prepare the content of the csv file.
    try:
        csv = pd.read_csv(
            fname,
            sep=";",
            index_col=1,
            parse_dates=[1],
            date_format="%Y%m%d%H",
        )
    except Exception:
        raise
    # We want to start at 2009 and go to the latest year.
    end_date = f"{year}-12-31 23:00:00"
    csv = csv.loc[start_date:end_date]
    # Some values might be missing to we fill them with the nearest
    # observations.
    index = pd.date_range(start=start_date, end=end_date, freq="h")
    try:
        csv = csv.reindex(index, method="nearest")
    except ValueError:
        # Something went wrong with indexing
        if len(index) == len(csv.index):
            csv.index = index
        else:
            csv = csv[~csv.index.duplicated()]
            csv.index = index
    except TypeError:
        csv.index = pd.DatetimeIndex(
            [datetime.strptime(x.split(":")[0], "%Y%m%d%H") for x in csv.index]
        )
        csv = csv[~csv.index.duplicated()]
        csv = csv.reindex(index, method="nearest")
    # Now we can copy the content of the csv to the dataframe
    for src_col, tar_col, fac in DATA_COLS[target]:
        csv[src_col] = csv[src_col].replace(-999, np.nan).interpolate()
        data[tar_col] = csv[src_col].values * fac

        if target == "air":
            # We need the day average air temperature for some of our
            # models.
            tar_col2 = f"day_avg_{tar_col}"
            data[tar_col2] = (
                csv[src_col].values.reshape(-1, 24).mean(axis=1).repeat(24)
            )
        if target == "solar":
            # Missing values are negative. We need to fill those
            missing = data[data[tar_col] < 0]
            data.loc[missing.index, tar_col] = np.nan
            data = data.interpolate()

    return data



"""MIDAS upgrade module for the weather data simulator."""

import logging
from importlib import import_module

import pandas as pd
from midas.scenario.upgrade_module import UpgradeModule
from midas.util.runtime_config import RuntimeConfig

from .analysis import analyze
from .download import download_weather
from .meta import (
    AVG_T_AIR,
    BI,
    DI,
    PRESSURE,
    RADIATION_DELTA,
    T_AIR,
    T_AIR_DELTA,
    WIND,
    WIND_DELTA,
    WINDDIR,
)

LOG = logging.getLogger(__name__)


class WeatherDataModule(UpgradeModule):
    def __init__(self):
        super().__init__(
            module_name="weather",
            default_scope_name="bremen",
            default_sim_config_name="WeatherData",
            default_import_str=(
                "midas_weather.simulator:WeatherDataSimulator"
            ),
            default_cmd_str=("%(python)s -m midas_weather.simulator %(addr)s"),
            log=LOG,
        )

        self.attrs = {
            T_AIR: [-20, 45],
            AVG_T_AIR: [-20, 45],
            BI: [0, 1368],
            DI: [0, 1368],
            WIND: [0, 150],
            WINDDIR: [0, 360],
            PRESSURE: [700, 1100],
            # wind direction
        }
        self.input_attrs = {
            WIND_DELTA: [-20, 20],
            T_AIR_DELTA: [-20, 20],
            RADIATION_DELTA: [-20, 20],
        }
        self.models = {
            "WeatherCurrent": [attr for attr in self.attrs],
            "WeatherForecast": [f"forecast_{attr}" for attr in self.attrs],
        }
        self._sensors: list = []
        self._actuators: list = []

    def check_module_params(self, module_params):
        """Check the module params and provide default values."""

        module_params.setdefault("start_date", self.scenario.base.start_date)
        module_params.setdefault("data_path", self.scenario.base.data_path)
        module_params.setdefault("interpolate", False)
        module_params.setdefault("noise_factor", 0.05)
        module_params.setdefault(
            "forecast_horizon_hours", self.scenario.base.forecast_horizon_hours
        )
        module_params.setdefault("forecast_error", 0.05)

        if self.scenario.base.no_rng:
            module_params["randomize"] = False
        else:
            module_params.setdefault("randomize", False)

    def check_sim_params(self, module_params):
        """Check the params for a certain simulator instance."""

        self.sim_params.setdefault("start_date", module_params["start_date"])
        self.sim_params.setdefault("data_path", module_params["data_path"])
        self.sim_params.setdefault("interpolate", module_params["interpolate"])
        self.sim_params.setdefault(
            "noise_factor", module_params["noise_factor"]
        )
        self.sim_params.setdefault(
            "forecast_horizon_hours", module_params["forecast_horizon_hours"]
        )
        self.sim_params.setdefault(
            "forecast_error", module_params["forecast_error"]
        )
        self.sim_params.setdefault("randomize", module_params["randomize"])
        self.sim_params.setdefault("seed_max", self.scenario.base.seed_max)
        self.sim_params.setdefault("seed", self.scenario.create_seed())
        self.sim_params.setdefault(
            "filename", RuntimeConfig().data["weather"][0]["name"]
        )

    def start_models(self):
        """Start models of a certain simulator."""
        mapping_key = "weather_mapping"
        if not self.sim_params.setdefault(
            mapping_key, self.create_default_mapping()
        ):
            # No mapping configured
            return

        if ":" in self.default_import_str:
            mod, clazz = self.default_import_str.split(":")
        else:
            mod, clazz = self.default_import_str.rsplit(".", 1)
        mod = import_module(mod)

        sim_dummy = getattr(mod, clazz)()
        sim_dummy.init(self.sid, **self.sim_params)

        for model, configs in self.sim_params[mapping_key].items():
            model_low = model.lower()

            for idx, config in enumerate(configs):
                model_key = self.scenario.generate_model_key(
                    self, model_low, idx
                )
                self.start_model(model_key, model, config)
                entity = sim_dummy.create(1, model, **config)[0]
                if model == "WeatherForecast":
                    # TODO: add as soon as it is requested
                    continue
                for attr, interval in self.attrs.items():
                    self._sensors.append(
                        {
                            "uid": f"{self.sid}.{entity['eid']}.{attr}",
                            "space": (
                                f"Box(low={interval[0]}, high={interval[1]}, "
                                "shape=(), dtype=np.float32)"
                            ),
                        }
                    )
                for attr, interval in self.input_attrs.items():
                    self._actuators.append(
                        {
                            "uid": f"{self.sid}.{entity['eid']}.{attr}",
                            "space": (
                                f"Box(low={interval[0]}, high={interval[1]}, "
                                "shape=(), dtype=np.float32)"
                            ),
                        }
                    )

    def connect(self):
        if self.sim_params["with_timesim"]:
            for model in self.models:
                if model not in self.sim_params["weather_mapping"]:
                    continue

                timesim, _ = self.scenario.find_first_model("timesim")
                for idx, _ in enumerate(
                    self.sim_params["weather_mapping"][model]
                ):
                    entity = self.scenario.generate_model_key(
                        self, model.lower(), idx
                    )
                    self.connect_entities(
                        timesim, entity, [("local_time", "now")]
                    )

    def connect_to_db(self):
        """Connect models to db."""
        mapping_key = "weather_mapping"
        db_key = self.scenario.find_first_model("store", "database")[0]

        for model, attrs in self.models.items():
            if model == "WeatherForecast":
                # TODO: add as soon as arrays/lists can be stored
                continue

            for idx, _ in enumerate(self.sim_params[mapping_key][model]):
                model_key = self.scenario.generate_model_key(
                    self, model.lower(), idx
                )
                self.connect_entities(model_key, db_key, attrs)

    def create_default_mapping(self):
        default_mapping = {}
        if not self.sim_params["weather_mapping"]:
            self.sim_params["weather_mapping"]["WeatherCurrent"] = [
                {
                    "interpolate": self.sim_params["interpolate"],
                    "randomize": self.sim_params["randomize"],
                }
            ]
        return default_mapping

    def download(self, data_path, tmp_path, force):
        download_weather(data_path, tmp_path, force)

    def analyze(
        self,
        name: str,
        data: pd.HDFStore,
        output_folder: str,
        start: int,
        end: int,
        step_size: int,
        full: bool,
    ):
        analyze(name, data, output_folder, start, end, step_size, full)

    def get_sensors(self):
        for sensor in self._sensors:
            self.scenario.sensors.append(sensor)
            LOG.debug(f"Created sensor entry {sensor}.")

        self._sensors = []

    def get_actuators(self):
        for actuator in self._actuators:
            self.scenario.actuators.append(actuator)
            LOG.debug(f"Created actuator entry {actuator}.")

        self._actuators = []



import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from midas.util import report_util

from .meta import (
    AVG_T_AIR,
    BI,
    CLOUDINESS,
    DI,
    PRESSURE,
    SUN_HOURS,
    T_AIR,
    WIND,
    WINDDIR,
)

ATTR_UNIT_MAP = {
    T_AIR: "temperature [°C]",
    AVG_T_AIR: "temperature [°C]",
    BI: "radiation [W/m²]",
    DI: "radiation [W/m²]",
    WIND: "speed [m/s]",
    WINDDIR: "direction [°]",
    PRESSURE: "pressure [hPA]",
    SUN_HOURS: "[min/h]",
    CLOUDINESS: "percentage [%]",
}


def analyze(
    name: str,
    data: pd.HDFStore,
    output_folder: str,
    start: int,
    end: int,
    step_size: int,
    full: bool,
):
    weather_sim_keys = [
        sim_key for sim_key in data.keys() if "WeatherData" in sim_key
    ]

    for sim_key in weather_sim_keys:
        wdata = data[sim_key]
        if start > 0:
            wdata = wdata.iloc[start:]
        if end > 0:
            wdata = wdata.iloc[:end]

        analyze_weather(
            wdata,
            step_size,
            f"{name}-{sim_key.replace('/', '')}",
            output_folder,
            full,
        )


def analyze_weather(data, step_size, name, output_path, full_report):
    plot_path = os.path.join(
        output_path, name.rsplit("-", 1)[1].replace("__", "_")
    )
    os.makedirs(plot_path, exist_ok=True)

    report_content = []

    report_path = os.path.join(output_path, f"{name}_report.md")
    report_file = open(report_path, "w")
    stats = {}
    for col in data.columns:
        _, attr = col.split("___")
        stats[attr] = {
            "Min": data[col].min(),
            "Max": data[col].max(),
            "Mean": data[col].mean(),
            "Std": data[col].std(),
        }
        stats[attr]["plot"] = _plot_series(data[col], attr, name, plot_path)
    _create_report(stats, report_content)

    for line in report_content:
        report_file.write(f"{line}\n")
    report_file.close()

    report_util.convert_markdown(report_path)


def _plot_series(data, attr, name, output_path):
    series = data.values

    annual = np.sort(series)[::-1]
    _, (ax1, ax2) = plt.subplots(2, 1, figsize=(12, 12))
    ax1.plot(series)
    ax1.set_title(f"{attr}")
    ax1.set_ylabel(ATTR_UNIT_MAP[attr])

    ax2.plot(annual)
    ax2.set_title(f"annual curve {attr}")
    ax2.set_ylabel(ATTR_UNIT_MAP[attr])
    ax2.set_xlabel("time [s]")

    filename = os.path.join(output_path, f"{name}_{attr}.png")
    plt.savefig(filename, dpi=300, bbox_inches="tight")
    plt.close()

    return filename


def _create_report(stats, report):
    report.append("# Analysis of Weather Data\n")

    for attr, keys in stats.items():
        report.append(
            f"## Summary of {attr}\n\n*Quantity [Unit]*: "
            f"{ATTR_UNIT_MAP[attr]}\n"
        )
        for key, val in keys.items():
            if key != "plot":
                report.append(f"- {key}: {val:.3f}")

        report.append(f"\n![{attr}]({keys['plot']})" + "{width=60%}\n")
    report.append("")



import os
import unittest

from midas.util.runtime_config import RuntimeConfig

from midas_weather.download import download_weather


class TestDownload(unittest.TestCase):
    def setUp(self):
        self.data_path = RuntimeConfig().paths["data_path"]
        self.tmp_path = os.path.join(self.data_path, "tmp")
        os.makedirs(self.tmp_path, exist_ok=True)

    @unittest.skip
    def test_download(self):
        download_weather(self.data_path, self.tmp_path, True)


if __name__ == "__main__":
    unittest.main()



"""This module contains the mosaik meta for the weather simulator."""

DATE = "Datetime"
T_AIR = "t_air_deg_celsius"
AVG_T_AIR = "day_avg_t_air_deg_celsius"
GHI = "gh_w_per_m2"
BI = "bh_w_per_m2"
DI = "dh_w_per_m2"
WIND = "wind_v_m_per_s"
WINDDIR = "wind_dir_deg"
PRESSURE = "air_pressure_hpa"
CLOUDINESS = "cloud_percent"
SUN_HOURS = "sun_hours_min_per_h"
T_AIR_DELTA = "delta_" + T_AIR
RADIATION_DELTA = "delta_" + GHI
WIND_DELTA = "delta_" + WIND


META = {
    "type": "time-based",
    "models": {
        "WeatherCurrent": {
            "public": True,
            "params": [
                "data_path",
                "start_date",
                "interpolate",
                "randomize",
                "seed",
                "block_mode",
                "frame",
            ],
            "attrs": [
                "now",
                BI,
                DI,
                T_AIR,
                AVG_T_AIR,
                WIND,
                WINDDIR,
                PRESSURE,
                CLOUDINESS,
                SUN_HOURS,
                T_AIR_DELTA,
                RADIATION_DELTA,
                WIND_DELTA,
            ],
        },
        "WeatherForecast": {
            "public": True,
            "params": [
                "data_path",
                "start_date",
                "interpolate",
                "forecast_error",
                "seed",
                "randomize",
                "forecast_horizon_hours",
                "block_mode",
                "frame",
            ],
            "attrs": [
                "now",
                "forecast_horizon_hours",
                f"forecast_{BI}",
                f"forecast_{DI}",
                f"forecast_{T_AIR}",
                f"forecast_{AVG_T_AIR}",
                f"forecast_{WIND}",
                f"forecast_{WINDDIR}",
                f"forecast_{PRESSURE}",
                f"forecast_{CLOUDINESS}",
                f"forecast_{SUN_HOURS}",
            ],
        },
    },
}



import os
import unittest

import pandas as pd
from midas.util.runtime_config import RuntimeConfig

from midas_weather.analysis import analyze


class TestDownload(unittest.TestCase):
    def setUp(self):
        self.output_path = os.path.join(
            RuntimeConfig().paths["output_path"], "midasmv_der"
        )
        # self.tmp_path = os.path.join(self.data_path, "tmp")
        self.db = pd.HDFStore(
            "/home/sbalduin/Code/Midas/midas-outputs/midasmv_der.hdf5"
        )
        # os.makedirs(self.tmp_path, exist_ok=True)

    @unittest.skip
    def test_analyze(self):
        analyze("Weather", self.db, self.output_path, 0, 0, 900, False)

    def tearDown(self):
        self.db.close()


if __name__ == "__main__":
    unittest.main()



"""This module contains the Weather simulator."""

import json
import logging
import os
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, Optional

import mosaik_api_v3
import numpy as np
import pandas as pd
from midas.util.dateformat import GER
from midas.util.logging import set_and_init_logger
from midas.util.print_format import mformat
from midas.util.runtime_config import RuntimeConfig

from .meta import META
from .model.current import WeatherCurrent
from .model.forecast import WeatherForecast
from .model.provider import WeatherData
from .module import WeatherDataModule

LOG = logging.getLogger("midas_weather.simulator")


class WeatherDataSimulator(mosaik_api_v3.Simulator):
    """The Weather simulator."""

    def __init__(self):
        super().__init__(META)
        self.sid: str
        self.models: Dict[str, Any] = dict()
        self.step_size: int
        self.now_dt: datetime
        self.rng: np.random.RandomState
        self.randomize: bool
        self.interpolate: bool
        self.forecast_horizon_hours: float
        self.forecast_error: float
        self.seed_max: int
        self._time_synced: bool = False
        self._sim_time: int = 0

    def init(
        self,
        sid: str,
        start_date: str,
        step_size: int = 900,
        data_path: str = RuntimeConfig().paths["data_path"],
        filename: str = RuntimeConfig().data["weather"][0]["name"],
        interpolate: bool = False,
        seed: Optional[int] = None,
        seed_max: int = 2**32 - 1,
        randomize: bool = False,
        forecast_horizon_hours: Optional[float] = None,
        forecast_error: float = 0.05,
        key_value_logs: Optional[bool] = None,
        **kwargs,
    ):
        """Called exactly ones after the simulator has been started.

        Parameters
        ----------
        sid : str
            Simulator ID for this simulator.
        start_date : str
            Start date as UTC ISO 8601 timestring such as
            '2019-01-01 00:00:00+0100'.
        step_size : int, optional
            Step size for this simulator. Defaults to 900.
        data_path : str, optional
            Path to the data folder. Defaults to the data folder in the
            midas root folder.
        filename : str, optional
            Name of the weather database. Defaults to
            *weather_bre2009-2019.hdf5*, the file that is created by
            the *midas.tools.weather_data.build_weather_data* function.
        interpolate : bool, optional
            If set to *True*, interpolation is enabled. Can be
            overwritten by model specific configuration.
        seed : int, optional
            A seed for the random number generator.
        randomize : bool, optional
            If set to *True*, randomization of data will be enabled.
            Otherwise, the data from the database is return unchanged.
            Can be overwritten by model specific configuration.

        Returns
        -------
        dict
            The meta dict (set by *mosaik_api.Simulator*).
        """

        self.sid = sid
        self.step_size = step_size
        self.now_dt = datetime.strptime(start_date, GER).astimezone(
            timezone.utc
        )
        self.rng = np.random.RandomState(seed)
        self.seed_max = seed_max
        self.interpolate = interpolate
        self.randomize = randomize
        if forecast_horizon_hours is None:
            self.forecast_horizon_hours = self.step_size * 2 / 3_600

        self.forecast_error = forecast_error

        self.wdata = self._load_data(data_path, filename)

        self.key_value_logs = key_value_logs
        if self.key_value_logs is None:
            self.key_value_logs = RuntimeConfig().misc.get(
                "key_value_logs", False
            )
        return self.meta

    def create(self, num, model, **model_params):
        """Initialize the simulation model instance (entity).

        Parameters
        ----------
        num : int
            The number of models to create.
        model : str
            The name of the models to create. Must be present inside
            the simulator's meta.

        Returns
        -------
        list
            A list with information on the created entity.
        """
        entities = list()

        for _ in range(num):
            eid = f"{model}-{len(self.models)}"
            seed = self.rng.randint(self.seed_max)

            if model == "WeatherCurrent":
                self.models[eid] = WeatherCurrent(
                    wdata=self.wdata,
                    start_date=self.now_dt,
                    step_size=self.step_size,
                    interpolate=model_params.get("interpolate", False),
                    randomize=model_params.get("randomize", False),
                    seed=model_params.get("seed", seed),
                )

            elif model == "WeatherForecast":
                self.models[eid] = WeatherForecast(
                    wdata=self.wdata,
                    start_date=self.now_dt,
                    step_size=self.step_size,
                    interpolate=model_params.get("interpolate", False),
                    randomize=model_params.get("randomize", False),
                    seed=model_params.get("seed", seed),
                    forecast_horizon_hours=model_params.get(
                        "forecast_horizon_hours", self.forecast_horizon_hours
                    ),
                    forecast_error=model_params.get(
                        "forecast_error", self.forecast_error
                    ),
                )
            entities.append({"eid": eid, "type": model})

        return entities

    def step(self, time: int, inputs: Dict[str, Any], max_advance: int = 0):
        """Perform a simulation step.

        Parameters
        ----------
        time : int
            The current simulation step (by convention in seconds since
            simulation start.
        inputs : dict
            A *dict* containing inputs for entities of this simulator.

        Returns
        -------
        int
            The next step this simulator wants to be stepped.

        """
        self._time_synced = False
        self._sim_time = time

        # Default inputs
        for model in self.models.values():
            model.step_size = self.step_size
            model.now_dt = self.now_dt

        # Inputs from other simulators
        for eid, attrs in inputs.items():
            log_msg = {
                "id": f"{self.sid}.{eid}",
                "name": eid,
                "type": eid.split("-")[1],
                "sim_time": self._sim_time,
                "msg_type": "input",
                "src_eids": [],
            }

            for attr, src_ids in attrs.items():
                setpoint = 0.0
                all_none = True
                for src_id, value in src_ids.items():
                    if value is not None:
                        all_none = False
                        if attr == "now":
                            attr = "now_dt"
                            setpoint = datetime.strptime(value, GER)
                            log_msg["src_eids"].append(src_id)
                            if not self._time_synced:
                                self.now_dt = setpoint
                                self._time_synced = True
                            break
                        setpoint += float(value)
                        log_msg["src_eids"].append(src_id)
                if not all_none:
                    log_msg[attr] = setpoint
                    setattr(self.models[eid], attr, setpoint)
            log_msg["src_eids"] = list(set(log_msg["src_eids"]))
            LOG.info(json.dumps(log_msg))

        # if inputs:
        #     if not self.key_value_logs:
        #         LOG.debug(
        #             "At step %d received inputs %s", time, mformat(inputs)
        #         )
        #     for eid, attrs in inputs.items():
        #         if self._time_synced:
        #             break
        #         for attr, src_ids in attrs.items():
        #             if self.key_value_logs:
        #                 self._log_input(eid, attr, src_ids)
        #             if attr != "now":
        #                 continue
        #             for val in src_ids.values():
        #                 self.now_dt = datetime.strptime(val, GER)
        #                 self._time_synced = True
        #                 break

        for model in self.models.values():
            # model.step_size = self.step_size
            # model.now_dt = self.now_dt
            model.step()

        if not self._time_synced:
            self.now_dt += timedelta(seconds=self.step_size)

        return time + self.step_size

    def get_data(self, outputs):
        """Return the requested output (if feasible).

        Parameters
        ----------
        outputs : dict
            A *dict* containing requested outputs of each entity.

        Returns
        -------
        dict
            A *dict* containing the values of the requested outputs.

        """

        data = dict()
        for eid, attrs in outputs.items():
            data[eid] = dict()
            model = eid.split("-")[0]
            for attr in attrs:
                if attr not in self.meta["models"][model]["attrs"]:
                    raise ValueError(f"Unknown output attribute {attr}")

                value = getattr(self.models[eid], attr)
                if isinstance(value, pd.DataFrame):
                    value = value.to_json()
                if not isinstance(value, str):
                    value = float(value)
                data[eid][attr] = value
                if self.key_value_logs:
                    self._log_output(eid, attr, value)
        if not self.key_value_logs:
            LOG.debug(
                "At step %d gathered ouputs %s", self._sim_time, mformat(data)
            )
        return data

    def _load_data(self, data_path, filename):
        filepath = os.path.join(data_path, filename)
        if not os.path.isfile(filepath):
            LOG.error(
                "File at '%s' does not exist. Will try to fix this...",
                filepath,
            )
            filepath = os.path.join(
                RuntimeConfig().paths["data_path"],
                RuntimeConfig().data["weather"][0]["name"],
            )
            if not os.path.isfile(filepath):
                try:
                    tmp_path = os.path.join(
                        RuntimeConfig().paths["data_path"], "tmp"
                    )
                    os.makedirs(tmp_path, exist_ok=True)
                    WeatherDataModule().download(
                        RuntimeConfig().paths["data_path"],
                        tmp_path,
                        True,
                        False,
                    )
                    assert os.path.isfile(filepath)

                except Exception:
                    LOG.exception(
                        "Unable to find data set at '%s'. Will die now!",
                        filepath,
                    )
        wdata = WeatherData(filepath, self.rng.randint(self.seed_max))
        return wdata

    def _log_input(self, eid: str, attr: str, src_ids: Dict[str, Any]):
        for src_id, val in src_ids.items():
            if isinstance(val, np.ndarray):
                val = val.tolist()
            elif isinstance(val, np.integer):
                val = int(val)
            elif isinstance(val, (np.float32, np.floating)):
                val = float(val)

            LOG.debug(
                json.dumps(
                    {
                        "id": f"{self.sid}_{eid}",
                        "name": eid,
                        "type": eid.split("-")[0],
                        "msg_type": "input",
                        "attribute": attr,
                        "source": src_id,
                        "value": val,
                    }
                )
            )

    def _log_output(self, eid: str, attr: str, value: Any):
        LOG.debug(
            json.dumps(
                {
                    "id": f"{self.sid}_{eid}",
                    "name": eid,
                    "model": eid.split("-")[0],
                    "type": "output",
                    "attribute": attr,
                    "value": value,
                }
            )
        )


if __name__ == "__main__":
    set_and_init_logger(
        0, "weather-logfile", "midas-weather.log", replace=True
    )
    LOG.info("Starting mosaik simulation...")
    mosaik_api_v3.start_simulation(WeatherDataSimulator())



"""This module contains the test for the weather-forecast model."""

import os
import unittest
from datetime import datetime, timezone
from os.path import abspath, join

from midas.util.runtime_config import RuntimeConfig

from midas_weather.download import download_weather
from midas_weather.model.forecast import WeatherForecast
from midas_weather.model.provider import WeatherData


class TestWeatherForecast(unittest.TestCase):
    """Test class for the weather-forecast model."""

    def setUp(self):
        data_path = RuntimeConfig().paths["data_path"]
        tmp_path = os.path.abspath(os.path.join(data_path, "tmp"))
        os.makedirs(tmp_path, exist_ok=True)

        download_weather(data_path, tmp_path, False)

        self.datapath = abspath(
            join(data_path, RuntimeConfig().data["weather"][0]["name"])
        )

        self.now_dt = datetime(
            year=2018,
            month=3,
            day=24,
            hour=22,
            minute=30,
            second=0,
            tzinfo=timezone.utc,
        )

    def test_init(self):
        """Test the init function."""
        weather = WeatherForecast(
            wdata=WeatherData(filename=self.datapath),
            start_date=self.now_dt,
            forecast_horizon_hours=3,
        )
        self.assertIsInstance(weather.wdata, WeatherData)
        self.assertIsInstance(weather.now_dt, datetime)

    def test_step(self):
        """Test the step function."""
        weather = WeatherForecast(
            wdata=WeatherData(filename=self.datapath),
            start_date=self.now_dt,
            step_size=900,
            forecast_horizon_hours=5,
            forecast_error=0,
        )

        weather.step()

        self.assertEqual(len(weather.t_air_deg_celsius), 20)
        self.assertEqual(len(weather.day_avg_t_air_deg_celsius), 20)
        self.assertEqual(len(weather.bh_w_per_m2), 20)
        self.assertEqual(len(weather.dh_w_per_m2), 20)
        self.assertEqual(len(weather.wind_v_m_per_s), 20)

        expected = [
            7.0,
            7.1,
            7.1,
            7.1,
            7.1,
            7.1,
            7.1,
            7.1,
            7.1,
            6.8,
            6.8,
            6.8,
            6.8,
            6.3,
            6.3,
            6.3,
            6.3,
            5.8,
            5.8,
            5.8,
        ]
        self.assertEqual(weather.t_air_deg_celsius, expected)

    def test_interpolate(self):
        """Test the step function with interpolate option."""
        weather = WeatherForecast(
            wdata=WeatherData(filename=self.datapath),
            start_date=self.now_dt,
            step_size=900,
            forecast_horizon_hours=4,
            forecast_error=0,
            interpolate=True,
        )

        weather.step()

        self.assertEqual(len(weather.t_air_deg_celsius), 16)
        self.assertEqual(len(weather.day_avg_t_air_deg_celsius), 16)
        self.assertEqual(len(weather.bh_w_per_m2), 16)
        self.assertEqual(len(weather.dh_w_per_m2), 16)

        expected = [
            [7.0, 7.1],
            [6.8, 7.1],
            [6.8, 7.1],
            [6.8, 7.1],
            [6.8, 7.1],
            [6.3, 6.8],
            [6.3, 6.8],
            [6.3, 6.8],
            [6.3, 6.8],
            [5.8, 6.3],
            [5.8, 6.3],
            [5.8, 6.3],
            [5.8, 6.3],
            [5.6, 6.8],
            [5.6, 6.8],
            [5.6, 6.8],
        ]
        for act, exp in zip(weather.t_air_deg_celsius, expected):
            self.assertTrue(exp[0] <= act <= exp[1])

    def test_with_error(self):
        """Test the step function with forecast_error option."""
        weather = WeatherForecast(
            wdata=WeatherData(filename=self.datapath),
            start_date=self.now_dt,
            step_size=900,
            forecast_horizon_hours=4,
            forecast_error=0.05,
            interpolate=False,
            seed=0,
            randomize=True,
        )

        weather.step()

        self.assertEqual(len(weather.t_air_deg_celsius), 16)
        self.assertEqual(len(weather.day_avg_t_air_deg_celsius), 16)
        self.assertEqual(len(weather.bh_w_per_m2), 16)
        self.assertEqual(len(weather.dh_w_per_m2), 16)

        expected = [
            7.0,
            7.1,
            7.1,
            7.1,
            7.1,
            6.8,
            6.8,
            6.8,
            6.8,
            6.3,
            6.3,
            6.3,
            6.3,
            5.8,
            5.8,
            5.8,
        ]
        self.assertNotEqual(weather.t_air_deg_celsius, expected)

    def test_forecast_returns_enough_values(self):
        for num in range(20):
            fch = num / 4
            weather = WeatherForecast(
                wdata=WeatherData(filename=self.datapath),
                start_date=self.now_dt,
                step_size=900,
                forecast_horizon_hours=fch,
                forecast_error=0.05,
                interpolate=False,
                seed=0,
                randomize=True,
            )

            weather.step()
            self.assertEqual(num, len(weather.forecast.index))


if __name__ == "__main__":
    unittest.main()



"""This module contains the forecast-weather model."""

import math
from datetime import timedelta

import numpy as np
import pandas as pd

from ..meta import (
    AVG_T_AIR,
    BI,
    CLOUDINESS,
    DI,
    PRESSURE,
    SUN_HOURS,
    T_AIR,
    WIND,
    WINDDIR,
)
from .current import WeatherCurrent
from .provider import COLS


class WeatherForecast(WeatherCurrent):
    """Weather forecast model.

    Uses the real data from future steps and applies a random error
    to the values.

    An internal start date is used and updated each step. This model
    provides weather information read from a weather time series.
    Additional functionality can be activated.

    Parameters
    ----------
    wdata : midas.core.weather.model.provider.WeatherData
        A reference to the class storing the actual data.
    start_date : datetime.datime
        The point of time in the year which should be started with.
    step_size : int, optional
        *step_size is an input and needs to be set in each step.*
        Used to step the internal datetime object. Does not reset
        after a step, so it can also be provided once during
        initialization.
    interpolate : bool, optional
        If set to *True*, data is interpolated if the internal datetime
        points to a time within two hours.
    seed : int, optional
        A seed for the random number generator.
    randomize : bool, optional
        If set to *True*, a normally distributed random noise is
        applied to the data.
    block_mode : bool, optional
        If set to *True*, the *select_block* function is used instead
        of the *select_hour* function. See the *WeatherData* to learn
        about the differences
    frame : int, optional
        If *block_mode* is True, this value is passed to the
        *select_block* function. Defaults to 1.
    forecast_horizon_hours : float, optional
        Specify for how many hours in the future a forecast should be
        generated.
    forecast_error : float, optional
        Specify how large the forecast error should be. Defaults to 0
        for a perfect forecast.

    Attributes
    ----------
    t_air_deg_celsius : float
        The current air temperature in degree celsius.
    day_avg_t_air_deg_celsius : float
        The current day's average air temperatur in degree celsius.
    bh_w_per_m2:
        Beam horizontal or direct solar radiation on the horizontal
        plane in Watt per square meter of the current step.
    dh_w_per_m2:
        Diffuse horizontal or diffuse solar radiation on the horizontal
        plane in Watt per square meter of the current step.

    """

    def __init__(
        self,
        wdata,
        start_date,
        step_size=None,
        forecast_horizon_hours=0.25,
        forecast_error=0,
        seed=None,
        interpolate=False,
        block_mode=False,
        randomize=False,
    ):
        super().__init__(
            wdata, start_date, step_size, randomize, interpolate, seed
        )

        self.forecast_horizon_hours = forecast_horizon_hours
        self.forecast_error = forecast_error
        self.forecast = None

        if block_mode:
            self.select_data = self.wdata.select_block
        else:
            self.select_data = self.wdata.select_hour

    def step(self):
        """Perform a simulation step."""

        pred_start = self.now_dt + timedelta(seconds=self.step_size)

        self.forecast = pd.DataFrame(
            columns=list(COLS.keys()),
            index=pd.date_range(
                pred_start,
                self.now_dt + timedelta(hours=self.forecast_horizon_hours),
                freq=f"{self.step_size}s",
            ),
        )
        horizon = int(math.ceil(self.forecast_horizon_hours)) + 2

        results = self.select_data(pred_start, horizon)

        steps = int(self.forecast_horizon_hours * 3_600 / self.step_size)

        if self.interpolate:
            val_idx = np.arange(0, (horizon + 1) * 3_600, 3_600)

            # start_hour = pred_start.hour
            start_sec = pred_start.minute * 60 + pred_start.second
            sel_idxs = []
            for _ in range(steps):
                sel_idxs.append(start_sec)
                start_sec += self.step_size

            results = [
                [np.interp(h, val_idx, results[idx]) for h in sel_idxs]
                for idx, _ in enumerate(COLS)
            ]

        else:
            sel_idxs = []
            start_hour = pred_start.hour
            for _ in range(steps):
                sel_idxs.append(pred_start.hour - start_hour)
                pred_start += timedelta(hours=self.step_size / 3_600)
                if pred_start.hour < start_hour:
                    start_hour = pred_start.hour - 1

            results = [
                [results[idx][h] for h in sel_idxs]
                for idx, _ in enumerate(COLS)
            ]

        if self.randomize:
            results = [
                [
                    v
                    * self.rng.normal(
                        scale=self.forecast_error * info[0], loc=1.0
                    )
                    for v in results[idx]
                ]
                for idx, info in enumerate(COLS.values())
            ]

        for idx, (col, info) in enumerate(COLS.items()):
            results[idx] = np.clip(
                results[idx], info[1][0], info[1][1]
            ).tolist()
            self.forecast[col] = results[idx]
            setattr(self, col, results[idx])

        return results

    @property
    def forecast_bh_w_per_m2(self):
        """Handle forecast_ - prefix."""
        return self.forecast.loc[:, [BI]]

    @property
    def forecast_dh_w_per_m2(self):
        """Handle forecast_ - prefix."""
        return self.forecast.loc[:, [DI]]

    @property
    def forecast_t_air_deg_celsius(self):
        """Handle forecast_ - prefix."""
        return self.forecast.loc[:, [T_AIR]]

    @property
    def forecast_day_avg_t_air_deg_celsius(self):
        """Handle forecast_ - prefix."""
        return self.forecast.loc[:, [AVG_T_AIR]]

    @property
    def forecast_wind_v_m_per_s(self):
        """Handle forecast_ - prefix."""
        return self.forecast.loc[:, [WIND]]

    @property
    def forecast_wind_dir_degree(self):
        """Handle forecast_ - prefix."""
        return self.forecast.loc[:, [WINDDIR]]

    @property
    def forecast_air_pressure_hpa(self):
        """Handle forecast_ - prefix."""
        return self.forecast.loc[:, [PRESSURE]]

    @property
    def forecast_sun_hours_min_per_h(self):
        """Handle forecast_ - prefix."""
        return self.forecast.loc[:, [SUN_HOURS]]

    @property
    def forecast_cloud_percent(self):
        """Handle forecast_ - prefix."""
        return self.forecast.loc[:, [CLOUDINESS]]



"""Test for the current-weather model."""

import os
import unittest
from datetime import datetime, timedelta, timezone
from os.path import abspath, join

from midas.util.runtime_config import RuntimeConfig

from midas_weather.download import download_weather
from midas_weather.model.current import WeatherCurrent
from midas_weather.model.provider import WeatherData


class TestWeatherCurrent(unittest.TestCase):
    """Test class for the current-weather model."""

    def setUp(self):
        data_path = RuntimeConfig().paths["data_path"]
        tmp_path = os.path.abspath(os.path.join(data_path, "tmp"))
        os.makedirs(tmp_path, exist_ok=True)

        download_weather(data_path, tmp_path, False)

        self.datapath = abspath(
            join(data_path, RuntimeConfig().data["weather"][0]["name"])
        )
        wdata = WeatherData(filename=self.datapath, seed=0)

        self.now_dt = datetime(
            year=2018,
            month=5,
            day=19,
            hour=13,
            minute=0,
            second=0,
            tzinfo=timezone.utc,
        )

        self.weather = WeatherCurrent(
            wdata=wdata, start_date=self.now_dt, seed=0
        )

    def test_init(self):
        """Test the init function."""
        # weather = WeatherCurrent(
        #     wdata=WeatherData(filename=self.datapath), start_date=self.now_dt
        # )
        self.assertIsInstance(self.weather.wdata, WeatherData)
        self.assertIsInstance(self.weather.now_dt, datetime)

    def test_step(self):
        """Test the step function."""
        self.weather.step()

        self.assertIsInstance(self.weather.t_air_deg_celsius, float)
        self.assertIsInstance(self.weather.day_avg_t_air_deg_celsius, float)
        self.assertIsInstance(self.weather.bh_w_per_m2, float)
        self.assertIsInstance(self.weather.dh_w_per_m2, float)
        self.assertIsInstance(self.weather.wind_v_m_per_s, float)
        self.assertIsInstance(self.weather.wind_dir_deg, float)
        self.assertIsInstance(self.weather.air_pressure_hpa, float)

        self.assertAlmostEqual(self.weather.t_air_deg_celsius, 17.2)
        self.assertAlmostEqual(
            self.weather.day_avg_t_air_deg_celsius, 12.316, places=2
        )
        self.assertAlmostEqual(self.weather.bh_w_per_m2, 33.334, places=2)
        self.assertAlmostEqual(self.weather.dh_w_per_m2, 416.667, places=2)
        self.assertAlmostEqual(self.weather.wind_v_m_per_s, 1.5)
        self.assertAlmostEqual(self.weather.wind_dir_deg, 120.0)
        self.assertAlmostEqual(self.weather.air_pressure_hpa, 1023.0)
        self.assertAlmostEqual(self.weather.sun_hours_min_per_h, 9.0)
        self.assertAlmostEqual(self.weather.cloud_percent, 75.0)

    def test_three_steps(self):
        """Test three calls of the step function.

        The first and second steps are 900s and, therefore, lower than
        the resolution of the dataset. The result will be the same.
        The third step a higher step size and the result will finally
        be differnt.
        """
        # First step
        self.weather.step()
        bh_w_per_m2 = self.weather.bh_w_per_m2
        dh_w_per_m2 = self.weather.dh_w_per_m2
        t_air_deg_celsius = self.weather.t_air_deg_celsius
        day_avg_t_air_deg_celsius = self.weather.day_avg_t_air_deg_celsius

        # Second step, no changes due to resolution
        self.weather.now_dt += timedelta(seconds=900)
        self.weather.step()
        self.assertEqual(t_air_deg_celsius, self.weather.t_air_deg_celsius)
        self.assertEqual(
            day_avg_t_air_deg_celsius, self.weather.day_avg_t_air_deg_celsius
        )
        self.assertEqual(bh_w_per_m2, self.weather.bh_w_per_m2)
        self.assertEqual(dh_w_per_m2, self.weather.dh_w_per_m2)

        self.weather.now_dt += timedelta(seconds=3600)
        self.weather.step()
        self.assertNotEqual(bh_w_per_m2, self.weather.bh_w_per_m2)
        self.assertNotEqual(dh_w_per_m2, self.weather.dh_w_per_m2)
        self.assertNotEqual(t_air_deg_celsius, self.weather.t_air_deg_celsius)
        self.assertEqual(
            day_avg_t_air_deg_celsius, self.weather.day_avg_t_air_deg_celsius
        )

    def test_interpolate(self):
        """Test interpolation of weather data."""
        self.weather.interpolate = True
        self.weather.auto_incr_time = True
        self.weather.step_size = 900

        t_airs = []
        self.weather.step()
        t_airs.append(self.weather.t_air_deg_celsius)
        self.weather.step()
        t_airs.append(self.weather.t_air_deg_celsius)
        self.weather.step()
        t_airs.append(self.weather.t_air_deg_celsius)
        self.weather.step()
        t_airs.append(self.weather.t_air_deg_celsius)
        self.weather.step()
        t_airs.append(self.weather.t_air_deg_celsius)

        for idx in range(1, len(t_airs)):
            self.assertGreater(t_airs[idx], t_airs[idx - 1])

    def test_seed(self):
        """Test randomization of weather data."""
        self.weather.interpolate = True
        self.weather.auto_incr_time = True
        self.weather.step_size = 900

        t_airs = []
        self.weather.step()
        t_airs.append(self.weather.t_air_deg_celsius)
        self.weather.step()
        t_airs.append(self.weather.t_air_deg_celsius)
        self.weather.step()
        t_airs.append(self.weather.t_air_deg_celsius)
        self.weather.step()
        t_airs.append(self.weather.t_air_deg_celsius)

        for idx in range(1, len(t_airs)):
            self.assertNotEqual(t_airs[idx], t_airs[idx - 1])


if __name__ == "__main__":
    unittest.main()



"""This module contains the test for the weather data provider."""

import os
import unittest
from datetime import datetime, timedelta, timezone
from os.path import abspath, join

from midas.util.dateformat import GER
from midas.util.runtime_config import RuntimeConfig
from pandas.core.frame import DataFrame

from midas_weather.download import download_weather
from midas_weather.meta import (
    AVG_T_AIR,
    CLOUDINESS,
    DI,
    GHI,
    PRESSURE,
    SUN_HOURS,
    T_AIR,
    WIND,
    WINDDIR,
)
from midas_weather.model.provider import WeatherData


class TestWeather(unittest.TestCase):
    """Test the weather data provider."""

    def setUp(self):
        data_path = RuntimeConfig().paths["data_path"]
        tmp_path = os.path.abspath(os.path.join(data_path, "tmp"))
        os.makedirs(tmp_path, exist_ok=True)

        download_weather(data_path, tmp_path, False)

        self.datapath = abspath(
            join(data_path, RuntimeConfig().data["weather"][0]["name"])
        )
        self.wdp = WeatherData(filename=self.datapath, seed=0)

    def test_init(self):
        """Test the init function."""
        self.assertIsInstance(self.wdp.wdata, DataFrame)
        cols = self.wdp.wdata.columns
        self.assertIn(AVG_T_AIR, cols)
        self.assertIn(T_AIR, cols)
        self.assertIn(GHI, cols)
        self.assertIn(DI, cols)
        self.assertIn(WIND, cols)
        self.assertIn(WINDDIR, cols)
        self.assertIn(PRESSURE, cols)
        self.assertIn(SUN_HOURS, cols)
        self.assertIn(CLOUDINESS, cols)

    def test_results(self):
        """Test the results."""

        now_dt = datetime.strptime("2018-05-19 12:00:00+0000", GER)
        data = self.wdp.select_hour(now_dt)

        self.assertEqual(data[0][0], 16.6)  # t_air
        self.assertAlmostEqual(data[1][0], 12.31666667)  # avg_t_air
        self.assertAlmostEqual(data[3][0], 386.1111111)  # diffuse radiation
        self.assertAlmostEqual(
            data[2][0], (411.1111111 - data[3][0])
        )  # direct radiation
        self.assertAlmostEqual(data[4][0], 1.3)  # wind speed
        self.assertAlmostEqual(data[5][0], 250)  # wind direction
        self.assertAlmostEqual(data[6][0], 1023.2)  # air pressure
        self.assertAlmostEqual(data[7][0], 12.0)  # sun hours
        self.assertAlmostEqual(data[8][0], 87.5)  # cloudiness

        data = self.wdp.select_hour(now_dt + timedelta(hours=1))
        self.assertEqual(data[0][0], 17.2)
        self.assertAlmostEqual(data[1][0], 12.316666667)
        self.assertAlmostEqual(data[3][0], 416.6666667)
        self.assertAlmostEqual(data[2][0], 450 - 416.6666667)
        self.assertAlmostEqual(data[4][0], 1.5)
        self.assertAlmostEqual(data[5][0], 120)
        self.assertAlmostEqual(data[6][0], 1023.0)
        self.assertAlmostEqual(data[7][0], 9.0)
        self.assertAlmostEqual(data[8][0], 75.0)

    def test_select_hour(self):
        """Test the select hour function."""
        now_dt = datetime.strptime("2017-12-31 23:00:00+0000", GER)

        data = self.wdp.select_hour(now_dt)

        self.assertIsInstance(data, list)
        self.assertEqual(len(data), 9)
        for datum in data:
            self.assertIsInstance(datum, list)
            self.assertEqual(len(datum), 1)
            for value in datum:
                self.assertIsInstance(value, float)

        # Access with different time zone should yield the same result
        now_dt = datetime.strptime("2018-01-01 00:00:00+0100", GER)
        data2 = self.wdp.select_hour(now_dt)

        for d1, d2 in zip(data, data2):
            self.assertEqual(d1[0], d2[0])

    @unittest.skip
    def test_select_hour_outside_dataset(self):
        """Test the datetime clipping to the dataset.

        Clipping only works into the future, not into the past.
        """
        data1 = self.wdp.select_hour(
            datetime.strptime("2025-01-01 00:00:00+0000", GER)
        )

        data2 = self.wdp.select_hour(
            datetime.strptime("2008-12-31 23:00:00+0000", GER)
        )

        for d1, d2 in zip(data1, data2):
            self.assertEqual(d1[0], d2[0])

        data3 = self.wdp.select_hour(
            datetime.strptime("2024-06-08 16:00:00+0200", GER)
        )
        # 2024 is a leap year, a direct mapping to 2010-06-08 is not
        # possible
        data4 = self.wdp.select_hour(
            datetime.strptime("2010-06-09 14:00:00+0000", GER)
        )

        for d1, d2 in zip(data3, data4):
            self.assertEqual(d1[0], d2[0])

        data5 = self.wdp.select_hour(
            datetime.strptime("2050-06-08 16:00:00+0200", GER)
        )
        data6 = self.wdp.select_hour(
            datetime.strptime("2022-06-09 14:00:00+0000", GER)
        )

        for d1, d2 in zip(data5, data6):
            self.assertEqual(d1[0], d2[0])

    def test_horizon(self):
        """The forecast horizon function.

        With a horizon of 2, the result should have three values.
        The requested current time plus the next two values.
        """
        now_dt = datetime.strptime("2017-12-31 23:00:00+0000", GER)

        data = self.wdp.select_hour(now_dt, horizon=2)
        self.assertIsInstance(data, list)
        self.assertEqual(len(data), 9)
        for datum in data:
            self.assertIsInstance(datum, list)
            self.assertEqual(len(datum), 3)

    def test_select_block(self):
        """Test the select block function."""
        now_dt = datetime.strptime("2018-01-01 00:00:00+0100", GER)

        data = self.wdp.select_block(now_dt, horizon=2, frame=1)
        self.assertIsInstance(data, list)
        self.assertEqual(len(data), 9)
        for datum in data:
            self.assertIsInstance(datum, list)
            self.assertEqual(len(datum), 3)

        # Temperature
        # 00:00 on 2017-12-31, 2018-01-01, 2018-01-02
        self.assertIn(data[0][0], [7.8, 9.1, 4.4])
        # 01:00 on 2017-12-31, 2018-01-01, 2018-01-02
        self.assertIn(data[0][1], [6.9, 7.9, 3.4])

    def test_select_block_year_change(self):
        """Test the select block function with a year change."""
        wdp = WeatherData(filename=self.datapath, seed=0)
        now_dt = datetime(
            year=2018,
            month=12,
            day=31,
            hour=23,
            minute=0,
            second=0,
            tzinfo=timezone(offset=timedelta(seconds=3600)),
        )
        for _ in range(10):  # test different random numbers
            data = wdp.select_block(now_dt, horizon=2, frame=1)
            self.assertIsInstance(data, list)
            self.assertEqual(len(data), 9)
            for datum in data:
                self.assertIsInstance(datum, list)
                self.assertEqual(len(datum), 3)
            self.assertIn(data[0][0], [6.7, 8.2, 4.6])
            self.assertIn(data[0][1], [6.7, 7.8, 5.1])
            self.assertIn(data[0][2], [6.7, 7.9, 6.1])


if __name__ == "__main__":
    unittest.main()



"""This module contains the provider for weather data."""

import logging
from datetime import timedelta, timezone

import numpy as np
import pandas as pd
from midas.util.dateformat import GER

from ..meta import (
    AVG_T_AIR,
    BI,
    CLOUDINESS,
    DI,
    GHI,
    PRESSURE,
    SUN_HOURS,
    T_AIR,
    WIND,
    WINDDIR,
)

COLS = {
    T_AIR: [1.42, [-20, 45.0]],
    AVG_T_AIR: [1.36, [-20, 45.0]],
    BI: [1.0, [0, 1368]],
    DI: [1.0, [0, 1368]],
    WIND: [1.0, [0, 150]],
    WINDDIR: [1.0, [0, 360]],
    PRESSURE: [0.0764, [700, 1100]],
    SUN_HOURS: [1.0, [0, 60]],
    CLOUDINESS: [1.0, [0, 100]],
}

LOG = logging.getLogger(__name__)


class WeatherData:
    """A simple weather data provider.

    This class provides weather data from a hdf file. Either a concrete
    datetime can be selected or randomly from a time frame.

    Parameters
    ----------
    filename : str
        Absolute path to the weather database.
    seed : int
        A seed for the randum number generatore. Random values are only
        used in the *select_block* method to pick randomly one value
        from the block. The values itself are unchanged.

    """

    def __init__(self, filename, seed=None):
        try:
            self.wdata = pd.read_csv(filename, index_col=0, date_format=GER)
        except (FileNotFoundError, UnicodeDecodeError):
            if filename.endswith(".hdf5"):
                self.wdata = pd.read_hdf(filename, "weather")
            else:
                LOG.critical(
                    "Weather data not available! Please update your Midas"
                    "runtime config and download the data. Run "
                    "'midasctl configure -u' followed by "
                    "'midasctl download -m weather' (without ')"
                )
                raise

        self.rng = np.random.RandomState(seed)

    def select_hour(self, now_dt, horizon=0):
        """
        Select weather data at given datetime *now_dt*.

        Returns
        -------
        list
            A *list* containing [[t_air], [avg_t_air], [GHI-DI], [DI]]

        """
        start_dt = now_dt.replace(
            minute=0, second=0, microsecond=0
        ).astimezone(timezone.utc)
        dates = [start_dt + timedelta(hours=h) for h in range(horizon + 1)]
        delta = (
            self.wdata.index[0].to_pydatetime()
            - self.wdata.index[-1].to_pydatetime()
            - timedelta(hours=1)
        )
        for idx in range(len(dates)):
            while dates[idx] > self.wdata.index[-1]:
                dates[idx] += delta

            dates[idx] = dates[idx].strftime(GER)

        data = self.wdata.loc[dates].copy()
        data[BI] = data[GHI] - data[DI]  # Direct radiation
        return [data[c].values.tolist() for c in COLS]

    def select_block(self, now_dt, horizon=0, frame=1):
        """Select weather data from a block related to *now_dt*.

        Gather the weather data from *frame* days in the past and in
        the future at exactly the same hour of the day and select
        randomly one of those values for each output. If self.horizon
        has a value > 1, then this is repeated for all hours in the
        horizon.

        Parameters
        ----------
        now_dt : datetime.datetime
            The time for which data is requested.
        horizon : int, optional
            Fetches data from the time steps following the provided
            time. Can be used to simulate weather forecast.
        frame : int, optional
            Specify the number of days which should be considered in
            each direction.

        """
        if frame < 1:
            return self.select_hour(now_dt, horizon)

        raw_data = []
        data = [[] for _ in COLS]

        base_dt = now_dt.replace(minute=0, second=0, microsecond=0).astimezone(
            timezone.utc
        )
        start_dt = base_dt - timedelta(days=frame)
        days = [start_dt + timedelta(days=day) for day in range(frame * 2 + 1)]

        for day in days:
            raw_data.append(self.select_hour(day, horizon))

        raw_data = np.array(raw_data)
        for day in range(frame * 2 + 1):
            for idx in range(len(COLS)):
                data[idx].append(self.rng.choice(raw_data[:, idx, day]))

        return data



"""This module contains the current-weather model."""

from datetime import timedelta

import numpy as np

from .provider import COLS


class WeatherCurrent:
    """This class provides weather information.

    An internal start date is used and updated each step. This model
    provides weather information read from a weather time series.
    Additional functionality can be activated.

    Parameters
    ----------
    wdata : midas.core.weather.model.provider.WeatherData
        A reference to the class storing the actual data.
    start_date : datetime.datime
        The point of time in the year which should be started with.
    step_size : int, optional
        *step_size is an input and needs to be set in each step.*
        Used to step the internal datetime object. Does not reset
        after a step, so it can also be provided once during
        initialization.
    interpolate : bool, optional
        If set to *True*, data is interpolated if the internal datetime
        points to a time within two hours.
    seed : int, optional
        A seed for the random number generator.
    randomize : bool, optional
        If set to *True*, a normally distributed random noise is
        applied to the data.
    block_mode : bool, optional
        If set to *True*, the *select_block* function is used instead
        of the *select_hour* function. See the *WeatherData* to learn
        about the differences
    frame : int, optional
        If *block_mode* is True, this value is passed to the
        *select_block* function. Defaults to 1.

    Attributes
    ----------
    t_air_deg_celsius : float
        The current air temperature in degree celsius.
    day_avg_t_air_deg_celsius : float
        The current day's average air temperatur in degree celsius.
    bh_w_per_m2:
        Beam horizontal or direct solar radiation on the horizontal
        plane in Watt per square meter of the current step.
    dh_w_per_m2:
        Diffuse horizontal or diffuse solar radiation on the horizontal
        plane in Watt per square meter of the current step.

    """

    def __init__(
        self,
        wdata,
        start_date,
        step_size=None,
        randomize=False,
        interpolate=False,
        seed=None,
        block_mode=False,
        frame=1,
    ):
        # Config
        self.wdata = wdata
        self.now_dt = start_date
        self.interpolate = interpolate
        self.randomize = randomize
        self.auto_incr_time = False

        # State
        self.t_air_deg_celsius = None
        self.day_avg_t_air_deg_celsius = None
        self.bh_w_per_m2 = None
        self.dh_w_per_m2 = None
        self.wind_v_m_per_s = None
        self.wind_dir_deg = None
        self.air_pressure_hpa = None
        self.sun_hours_min_per_h = None
        self.cloud_percent = None
        self._delta_t_air_deg_celsius = 0.0
        self._delta_gh_w_per_m2 = 0.0
        self._delta_wind_v_m_per_s = 0.0
        self.rng = np.random.RandomState(seed)

        # Input
        self.step_size = step_size
        self.block_mode = block_mode
        self.frame = frame
        self.delta_t_air_deg_celsius = 0.0
        self.delta_gh_w_per_m2 = 0.0
        self.delta_wind_v_m_per_s = 0.0

    def step(self):
        """Perform a simulation step."""

        select_dt = self.now_dt.replace(minute=0, second=0, microsecond=0)

        if self.block_mode:
            res = self.wdata.select_block(select_dt, 0, self.frame)
        else:
            res = self.wdata.select_hour(select_dt)
        if self.interpolate:
            select_dt = select_dt + timedelta(hours=1)
            if self.block_mode:
                res2 = self.wdata.select_block(select_dt, 1, self.frame)
            else:
                res2 = self.wdata.select_hour(select_dt)

            cur_sec = self.now_dt.minute * 60
            res = [
                [np.interp(cur_sec, [0, 3_600], [res[i][0], res2[i][0]])]
                for i in range(len(res))
            ]

        if self.randomize:
            res = [[r[0] * self.rng.normal(scale=0.05, loc=1.0)] for r in res]

        self._apply_delta(res)

        for idx, (col, info) in enumerate(COLS.items()):
            # Apply "physical" boundaries
            res[idx][0] = min(info[1][1], max(info[1][0], res[idx][0]))
            setattr(self, col, res[idx][0])

        # # Radiation can't be lower than 0
        # res[2][0] = max(0, res[2][0])
        # res[3][0] = max(0, res[3][0])
        # res[5][0] = max(0, min(360, res[5][0]))

        # print(res)
        # self.t_air_deg_celsius = res[0][0]
        # self.day_avg_t_air_deg_celsius = res[1][0]
        # self.bh_w_per_m2 = res[2][0]
        # self.dh_w_per_m2 = res[3][0]
        # self.wind_v_m_per_s = res[4][0]
        # self.wind_dir_deg = res[5][0]
        # self.air_pressure_hpa = res[6][0]
        # self.sun_hours_min_per_h = res[7][0]
        # self.cloud_percent = res[8][0]

        if self.auto_incr_time:
            self.now_dt += timedelta(seconds=self.step_size)

    def _apply_delta(self, res):
        self._delta_t_air_deg_celsius += self.delta_t_air_deg_celsius
        self._delta_gh_w_per_m2 += self.delta_gh_w_per_m2
        self._delta_wind_v_m_per_s += self.delta_wind_v_m_per_s

        self._delta_t_air_deg_celsius = min(
            100, max(-100, self._delta_t_air_deg_celsius)
        )
        self._delta_gh_w_per_m2 = min(100, max(-100, self._delta_gh_w_per_m2))
        self._delta_wind_v_m_per_s = min(
            100, max(-100, self._delta_wind_v_m_per_s)
        )

        self.delta_t_air_deg_celsius = 0
        self.delta_gh_w_per_m2 = 0
        self.delta_wind_v_m_per_s = 0

        res[0][0] += res[0][0] * self._delta_t_air_deg_celsius / 100
        res[1][0] += res[1][0] * self._delta_t_air_deg_celsius / 100
        res[2][0] += res[2][0] * self._delta_gh_w_per_m2 / 100
        res[3][0] += res[3][0] * self._delta_gh_w_per_m2 / 100
        res[4][0] += res[4][0] * self._delta_wind_v_m_per_s / 100



"""Test module for the weather simulator."""

import os
import unittest
from datetime import datetime, timedelta
from io import StringIO

import pandas as pd
from midas.util.dateformat import GER
from midas.util.runtime_config import RuntimeConfig

from midas_weather.download import download_weather
from midas_weather.model.provider import WeatherData
from midas_weather.simulator import WeatherDataSimulator


class TestSimulator(unittest.TestCase):
    """Test class for the weather simulator."""

    def setUp(self):
        data_path = RuntimeConfig().paths["data_path"]
        tmp_path = os.path.abspath(os.path.join(data_path, "tmp"))
        os.makedirs(tmp_path, exist_ok=True)

        download_weather(data_path, tmp_path, False)

        # self.sim = WeatherSimulator()
        self.params = {
            "sid": "TestSimulator-0",
            "step_size": 900,
            "data_path": data_path,
            "start_date": "2018-05-19 14:00:00+0100",
        }

    def test_init(self):
        """Test the init function of the simulator."""

        sim = WeatherDataSimulator()
        sim.init(**self.params)
        meta = sim.meta

        self.assertIsInstance(meta, dict)
        self.assertIsInstance(sim.wdata, WeatherData)
        self.assertIsInstance(sim.now_dt, datetime)

    def test_create_weather(self):
        """Test to create a weather model."""
        sim = WeatherDataSimulator()
        sim.init(**self.params)

        entities = sim.create(num=1, model="Weather", interpolate=False)
        self.assertEqual(len(entities), 1)
        for entity in entities:
            self.assertIsInstance(entity, dict)

    def test_create_weather_forecast(self):
        """Test to create a weather forecast model."""
        sim = WeatherDataSimulator()
        sim.init(**self.params)

        entities = sim.create(
            num=1,
            model="WeatherForecast",
            forecast_horizon_hours=12,
            forecast_error=0,
            seed=0,
            interpolate=False,
        )
        self.assertEqual(len(entities), 1)
        for entity in entities:
            self.assertIsInstance(entity, dict)

    def test_step(self):
        """Test a simulation step with both a weather model and a
        weather forecast model.

        """
        sim = WeatherDataSimulator()
        sim.init(**self.params)

        sim.create(1, "WeatherCurrent", interpolate=False)
        sim.create(
            1,
            "WeatherForecast",
            forecast_horizon_hours=12,
            forecast_error=0,
            interpolate=False,
            seed=0,
        )

        sim.step(0, {})

        t_air = sim.models["WeatherCurrent-0"].t_air_deg_celsius
        self.assertIsInstance(t_air, float)

        fc_t_air = sim.models["WeatherForecast-1"].forecast_t_air_deg_celsius
        self.assertIsInstance(fc_t_air, pd.DataFrame)

        self.assertEqual(t_air, fc_t_air.iloc[0]["t_air_deg_celsius"])

    def test_get_data(self):
        """Test the get data function of the simulator."""
        sim = WeatherDataSimulator()
        sim.init(**self.params)

        sim.create(1, "WeatherCurrent", interpolate=False)
        sim.create(
            1,
            "WeatherForecast",
            forecast_horizon_hours=12,
            forecast_error=0,
            interpolate=False,
            seed=0,
        )
        sim.step(0, {})

        outputs = {
            "WeatherCurrent-0": [
                "t_air_deg_celsius",
                "day_avg_t_air_deg_celsius",
            ],
            "WeatherForecast-1": [
                "forecast_bh_w_per_m2",
                "forecast_dh_w_per_m2",
            ],
        }

        data = sim.get_data(outputs)

        self.assertIsInstance(data, dict)
        self.assertEqual(len(data), 2)
        for key, value in data.items():
            self.assertIsInstance(key, str)
            self.assertIsInstance(value, dict)

            for attr, val in value.items():
                self.assertIsInstance(attr, str)
                if "forecast" in attr:
                    self.assertIsInstance(val, str)
                    val = pd.read_json(StringIO(val)).tz_localize("UTC")
                    self.assertEqual(val.size, 48)
                    self.assertIn(val.columns[0], attr)
                else:
                    self.assertIsInstance(val, float)

    @unittest.skip
    def test_complete_dataset(self):
        """Test if the whole dataset is usable.

        Takes very long, therefore disabled by default.
        """
        sim = WeatherDataSimulator()
        params = {
            "sid": "TestSimulator-0",
            "step_size": 3600,
            "data_path": RuntimeConfig().paths["data_path"],
            "start_date": "2008-12-31 23:00:00+0000",
        }
        sim.init(**params)
        sim.create(1, "WeatherCurrent", interpolate=False, randomize=False)
        sim.create(
            1,
            "WeatherForecast",
            interpolate=False,
            randomize=False,
            forecast_horizon_hours=2,
            forecast_error=0,
        )
        now_dt = datetime.strptime(params["start_date"], "%Y-%m-%d %H:%M:%S%z")
        years = ([8760, 8760, 8760, 8784] * 3)[:-1]
        time = 0
        wdata = sim.models["WeatherCurrent-0"].wdata.wdata
        for hours in years:
            for hour in range(hours):
                sim.step(time, dict())
                data = sim.get_data(
                    {
                        "WeatherCurrent-0": [
                            "t_air_deg_celsius",
                            "dh_w_per_m2",
                        ],
                        "WeatherForecast-1": ["forecast_t_air_deg_celsius"],
                    }
                )
                now_str = now_dt.strftime(GER)
                self.assertEqual(
                    data["WeatherCurrent-0"]["t_air_deg_celsius"],
                    wdata["t_air_degree_celsius"].loc[now_str],
                )
                try:
                    self.assertEqual(
                        data["WeatherCurrent-0"]["dh_w_per_m2"],
                        wdata["dh_w_per_m2"].loc[now_str],
                    )
                except AssertionError as err:
                    print()
                    raise err

                fc_start = (now_dt + timedelta(hours=1)).strftime(GER)
                fc_end = (now_dt + timedelta(hours=2)).strftime(GER)
                forecast = pd.read_json(
                    data["WeatherForecast-1"]["forecast_t_air_deg_celsius"]
                ).tz_localize("UTC")
                for val1, val2 in zip(
                    forecast["t_air_deg_celsius"].values,
                    wdata["t_air_degree_celsius"].loc[fc_start:fc_end].values,
                ):
                    self.assertAlmostEqual(val1, val2)

                time += 3600
                now_dt += timedelta(seconds=3600)
            print(sim.models["WeatherCurrent-0"].now_dt)
        print(hours)
        # print(len(years))

    def test_set_delta_input(self):
        sim = WeatherDataSimulator()
        sim.init(**self.params)

        sim.create(1, "WeatherCurrent", interpolate=False)

        inputs = {
            "WeatherCurrent-0": {
                "delta_t_air_deg_celsius": {"DummySim-0.DummyMod-0": -10},
                "delta_gh_w_per_m2": {"DummySim-0.DummyMod-0": 20},
                "delta_wind_v_m_per_s": {"DummySim-0.DummyMod-0": 0},
            }
        }
        sim.step(0, inputs)


if __name__ == "__main__":
    unittest.main()
